!MNH_LIC Copyright 1994-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
MODULE MODI_ICE4_RRHONG
INTERFACE
SUBROUTINE ICE4_RRHONG(LDSOFT, LDCOMPUTE, &
                       &PEXN, PLVFACT, PLSFACT, &
                       &PT,   PRRT, &
                       &PTHT, &
                       &PRRHONG_MR, PB_TH, PB_RR, PB_RG)
IMPLICIT NONE
LOGICAL,                  INTENT(IN)    :: LDSOFT
LOGICAL, DIMENSION(:),    INTENT(IN)    :: LDCOMPUTE
REAL, DIMENSION(:),       INTENT(IN)    :: PEXN     ! Exner function
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT  ! L_v/(Pi_ref*C_ph)
REAL, DIMENSION(:),       INTENT(IN)    :: PLSFACT  ! L_s/(Pi_ref*C_ph)
REAL, DIMENSION(:),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(:),       INTENT(IN)    :: PRRT     ! Rain water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PTHT     ! Theta at t
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRHONG_MR ! Mixing ratio change due to spontaneous freezing
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_RR
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_RG
END SUBROUTINE ICE4_RRHONG
END INTERFACE
END MODULE MODI_ICE4_RRHONG
SUBROUTINE ICE4_RRHONG(LDSOFT, LDCOMPUTE, &
                       &PEXN, PLVFACT, PLSFACT, &
                       &PT,   PRRT, &
                       &PTHT, &
                       &PRRHONG_MR, PB_TH, PB_RR, PB_RG)
!!
!!**  PURPOSE
!!    -------
!!      Computes the RRHONG process
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the splitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!!
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,            ONLY: XTT
USE MODD_RAIN_ICE_DESCR, ONLY: XRTMIN
USE MODD_PARAM_ICE,      ONLY: LFEEDBACKT
!
USE MODE_MPPDB
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
LOGICAL,                  INTENT(IN)    :: LDSOFT
LOGICAL, DIMENSION(:),    INTENT(IN)    :: LDCOMPUTE
REAL, DIMENSION(:),       INTENT(IN)    :: PEXN     ! Exner function
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT  ! L_v/(Pi_ref*C_ph)
REAL, DIMENSION(:),       INTENT(IN)    :: PLSFACT  ! L_s/(Pi_ref*C_ph)
REAL, DIMENSION(:),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(:),       INTENT(IN)    :: PRRT     ! Rain water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PTHT     ! Theta at t
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRHONG_MR ! Mixing ratio change due to spontaneous freezing
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_RR
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_RG
!
!*       0.2  declaration of local variables
!
LOGICAL, DIMENSION(:), allocatable :: GMASK
!
!$acc data present(LDCOMPUTE,PEXN,PLVFACT,PLSFACT,PT,PRRT,PTHT, &
!$acc&                PRRHONG_MR,PB_TH,PB_RR,PB_RG)
!-------------------------------------------------------------------------------
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PEXN,"ICE4_RRHONG beg:PEXN")
  CALL MPPDB_CHECK(PLVFACT,"ICE4_RRHONG beg:PLVFACT")
  CALL MPPDB_CHECK(PLSFACT,"ICE4_RRHONG beg:PLSFACT")
  CALL MPPDB_CHECK(PT,"ICE4_RRHONG beg:PT")
  CALL MPPDB_CHECK(PRRT,"ICE4_RRHONG beg:PRRT")
  CALL MPPDB_CHECK(PTHT,"ICE4_RRHONG beg:PTHT")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PB_TH,"ICE4_RRHONG beg:PB_TH")
  CALL MPPDB_CHECK(PB_RR,"ICE4_RRHONG beg:PB_RR")
  CALL MPPDB_CHECK(PB_RG,"ICE4_RRHONG beg:PB_RG")
END IF

allocate( gmask(size( prrt ) ) )

!$acc data create(GMASK)
!
!*       3.3     compute the spontaneous freezing source: RRHONG
!
!$acc kernels
PRRHONG_MR(:) = 0.
IF(LDSOFT) THEN
  GMASK(:)=PT(:)<XTT-35.0 .AND. PRRT(:)>XRTMIN(3) .AND. LDCOMPUTE(:)
  WHERE(GMASK(:))
    PRRHONG_MR(:) = PRRT(:)
  ENDWHERE
  IF(LFEEDBACKT) THEN
    !Limitation due to -35 crossing of temperature
    WHERE(GMASK(:))
      PRRHONG_MR(:)=MIN(PRRHONG_MR(:), MAX(0., ((XTT-35.)/PEXN(:)-PTHT)/(PLSFACT(:)-PLVFACT(:))))
    ENDWHERE
  ENDIF
ENDIF
PB_RG(:) = PB_RG(:) + PRRHONG_MR(:)
PB_RR(:) = PB_RR(:) - PRRHONG_MR(:)
PB_TH(:) = PB_TH(:) + PRRHONG_MR(:)*(PLSFACT(:)-PLVFACT(:))
!$acc end kernels
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PB_TH,"ICE4_RRHONG end:PB_TH")
  CALL MPPDB_CHECK(PB_RR,"ICE4_RRHONG end:PB_RR")
  CALL MPPDB_CHECK(PB_RG,"ICE4_RRHONG end:PB_RG")
  !Check all OUT arrays
  CALL MPPDB_CHECK(PRRHONG_MR,"ICE4_RRHONG end:PRRHONG_MR")
END IF

!$acc end data

!$acc end data

END SUBROUTINE ICE4_RRHONG
