!MNH_LIC Copyright 1995-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
! Modifications:
!  P. Wautelet 25/02/2019: split rain_ice (cleaner and easier to maintain/debug)
!  P. Wautelet 05/06/2019: optimisations
!-----------------------------------------------------------------
MODULE MODE_RAIN_ICE_FAST_RI

  IMPLICIT NONE

  PRIVATE

  PUBLIC :: RAIN_ICE_FAST_RI

CONTAINS

SUBROUTINE RAIN_ICE_FAST_RI(OMICRO, PRHODREF, PRIT, PRHODJ, PZT, PSSI, PLSFACT, PLVFACT, &
                            PAI, PCJ, PRHODJ3D, PTHS3D, PCIT, PRCS, PRIS, PTHS)
!
!*      0. DECLARATIONS
!          ------------
!
use MODD_BUDGET,         only: LBUDGET_RC, LBUDGET_RI, LBUDGET_TH
use MODD_CST,            only: XTT
use MODD_RAIN_ICE_DESCR, only: XDI, XLBEXI, XLBI, XRTMIN
use MODD_RAIN_ICE_PARAM, only: X0DEPI, X2DEPI
!
use mode_mppdb
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
use MODI_BUDGET
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
LOGICAL,  DIMENSION(:,:,:), intent(in)    :: OMICRO   ! Test where to compute all processes
REAL,     DIMENSION(:),     intent(in)    :: PRHODREF ! RHO Dry REFerence
REAL,     DIMENSION(:),     intent(in)    :: PRIT     ! Pristine ice m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRHODJ   ! RHO times Jacobian
REAL,     DIMENSION(:),     intent(in)    :: PZT      ! Temperature
REAL,     DIMENSION(:),     intent(in)    :: PSSI     ! Supersaturation over ice
REAL,     DIMENSION(:),     intent(in)    :: PLSFACT  ! L_s/(Pi_ref*C_ph)
REAL,     DIMENSION(:),     intent(in)    :: PLVFACT  ! L_v/(Pi_ref*C_ph)
REAL,     DIMENSION(:),     intent(in)    :: PAI      ! Thermodynamical function
REAL,     DIMENSION(:),     intent(in)    :: PCJ      ! Function to compute the ventilation coefficient
REAL,     DIMENSION(:,:,:), INTENT(IN)    :: PRHODJ3D ! Dry density * Jacobian
REAL,     DIMENSION(:,:,:), INTENT(IN)    :: PTHS3D   ! Theta source
REAL,     DIMENSION(:),     intent(inout) :: PCIT     ! Pristine ice conc. at t
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRCS     ! Cloud water m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRIS     ! Pristine ice m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PTHS     ! Theta source
!
!*       0.2  declaration of local variables
!
LOGICAL, DIMENSION(:), ALLOCATABLE :: GWORK
REAL,    DIMENSION(:), ALLOCATABLE :: ZZW  ! Work array
!
!-------------------------------------------------------------------------------
!
! IN variables
!
!$acc data present( OMICRO, PRHODREF, PRIT,              &
!$acc &             PRHODJ, PZT, PSSI, PLSFACT, PLVFACT, &
!$acc &             PAI, PCJ, PRHODJ3D, PTHS3D,          &
!
! INOUT variables
!
!$acc &             PCIT, PRCS, PRIS, PTHS )
!
! OUT variables
!
!NONE

IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(OMICRO,"RAIN_ICE_FAST_RI beg:OMICRO")
  CALL MPPDB_CHECK(PRHODREF,"RAIN_ICE_FAST_RI beg:PRHODREF")
  CALL MPPDB_CHECK(PRIT,"RAIN_ICE_FAST_RI beg:PRIT")
  CALL MPPDB_CHECK(PRHODJ,"RAIN_ICE_FAST_RI beg:PRHODJ")
  CALL MPPDB_CHECK(PZT,"RAIN_ICE_FAST_RI beg:PZT")
  CALL MPPDB_CHECK(PSSI,"RAIN_ICE_FAST_RI beg:PSSI")
  CALL MPPDB_CHECK(PLSFACT,"RAIN_ICE_FAST_RI beg:PLSFACT")
  CALL MPPDB_CHECK(PLVFACT,"RAIN_ICE_FAST_RI beg:PLVFACT")
  CALL MPPDB_CHECK(PAI,"RAIN_ICE_FAST_RI beg:PAI")
  CALL MPPDB_CHECK(PCJ,"RAIN_ICE_FAST_RI beg:PCJ")
  CALL MPPDB_CHECK(PRHODJ3D,"RAIN_ICE_FAST_RI beg:PRHODJ3D")
  CALL MPPDB_CHECK(PTHS3D,"RAIN_ICE_FAST_RI beg:PTHS3D")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PCIT,"RAIN_ICE_FAST_RI beg:PCIT")
  CALL MPPDB_CHECK(PRCS,"RAIN_ICE_FAST_RI beg:PRCS")
  CALL MPPDB_CHECK(PRIS,"RAIN_ICE_FAST_RI beg:PRIS")
  CALL MPPDB_CHECK(PTHS,"RAIN_ICE_FAST_RI beg:PTHS")
END IF
!
ALLOCATE( GWORK(size(PRHODREF)) )
ALLOCATE( ZZW  (size(PRHODREF)) )

!$acc data create( GWORK, ZZW )

!
!*       7.1    cloud ice melting
!
!$acc kernels
  GWORK(:) = PRIS(:)>0.0 .AND. PZT(:)>XTT
  WHERE( GWORK(:) )
    PRCS(:) = PRCS(:) + PRIS(:)
    PTHS(:) = PTHS(:) - PRIS(:)*(PLSFACT(:)-PLVFACT(:)) ! f(L_f*(-RIMLTC))
    PRIS(:) = 0.0
    PCIT(:) = 0.0
  END WHERE
!$acc end kernels
  IF (LBUDGET_TH) THEN
!$acc update self(PTHS)
    CALL BUDGET (UNPACK(PTHS(:),MASK=OMICRO(:,:,:),FIELD=PTHS3D)*PRHODJ3D(:,:,:),4,'IMLT_BU_RTH')
  END IF
  IF (LBUDGET_RC) THEN
!$acc update self(PRCS)
    CALL BUDGET (UNPACK(PRCS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),7,'IMLT_BU_RRC')
  END IF
  IF (LBUDGET_RI) THEN
!$acc update self(PRIS)
    CALL BUDGET (UNPACK(PRIS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),9,'IMLT_BU_RRI')
  END IF
!
!*       7.2    Bergeron-Findeisen effect: RCBERI
!
!$acc kernels
  GWORK(:) = PRCS(:)>0.0 .AND. PSSI(:)>0.0 .AND. PRIT(:)>XRTMIN(4) .AND. PCIT(:)>0.0
  WHERE( GWORK(:) )
#ifndef MNH_BITREP
    ZZW(:) = MIN(1.E8,XLBI*( PRHODREF(:)*PRIT(:)/PCIT(:) )**XLBEXI) ! Lbda_i
    ZZW(:) = MIN( PRCS(:),( PSSI(:) / (PRHODREF(:)*PAI(:)) ) * PCIT(:) * &
                  ( X0DEPI/ZZW(:) + X2DEPI*PCJ(:)*PCJ(:)/ZZW(:)**(XDI+2.0) ) )
#else
    ZZW(:) = MIN(1.E8,XLBI*BR_POW( PRHODREF(:)*PRIT(:)/PCIT(:), XLBEXI ) ) ! Lbda_i
    ZZW(:) = MIN( PRCS(:),( PSSI(:) / (PRHODREF(:)*PAI(:)) ) * PCIT(:) * &
                  ( X0DEPI/ZZW(:) + X2DEPI*PCJ(:)*PCJ(:)/BR_POW(ZZW(:),XDI+2.0) ) )
#endif
    PRCS(:) = PRCS(:) - ZZW(:)
    PRIS(:) = PRIS(:) + ZZW(:)
    PTHS(:) = PTHS(:) + ZZW(:)*(PLSFACT(:)-PLVFACT(:)) ! f(L_f*(RCBERI))
  END WHERE
!$acc end kernels
  IF (LBUDGET_TH) THEN
!$acc update self(PTHS)
    CALL BUDGET (UNPACK(PTHS(:),MASK=OMICRO(:,:,:),FIELD=PTHS3D)*PRHODJ3D(:,:,:),4,'BERFI_BU_RTH')
  END IF
  IF (LBUDGET_RC) THEN
!$acc update self(PRCS)
    CALL BUDGET (UNPACK(PRCS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),7,'BERFI_BU_RRC')
  END IF
  IF (LBUDGET_RI) THEN
!$acc update self(PRIS)
    CALL BUDGET (UNPACK(PRIS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),9,'BERFI_BU_RRI')
  END IF
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PCIT,"RAIN_ICE_FAST_RI end:PCIT")
  CALL MPPDB_CHECK(PRCS,"RAIN_ICE_FAST_RI end:PRCS")
  CALL MPPDB_CHECK(PRIS,"RAIN_ICE_FAST_RI end:PRIS")
  CALL MPPDB_CHECK(PTHS,"RAIN_ICE_FAST_RI end:PTHS")
END IF

!$acc end data

!$acc end data

END SUBROUTINE RAIN_ICE_FAST_RI

END MODULE MODE_RAIN_ICE_FAST_RI
