!MNH_LIC Copyright 1994-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!     ######################
      MODULE MODI_SLOW_TERMS
!     ###################### 
INTERFACE
      SUBROUTINE SLOW_TERMS ( KSPLITR, PTSTEP, KMI, HSUBG_AUCV,            &
                              PZZ, PRHODJ, PRHODREF, PCLDFR,               &
                              PTHT, PRVT, PRCT, PRRT, PPABST,              &
                              PTHS, PRVS, PRCS, PRRS, PINPRR,              &
                              PINPRR3D, PEVAP3D                            )
!
!
INTEGER,                  INTENT(IN)    :: KSPLITR ! Number of small time step 
                                      ! integration for  rain sedimendation
REAL,                     INTENT(IN)    :: PTSTEP  ! Double Time step
                                                   ! (single if cold start)
INTEGER,                  INTENT(IN)    :: KMI     ! Model index
CHARACTER(LEN=4),         INTENT(IN)   :: HSUBG_AUCV
                                        ! Kind of Subgrid autoconversion method
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PZZ     ! Height (z)
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHODJ  ! Dry density * Jacobian
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHODREF! Reference density
REAL, DIMENSION(:,:,:),     INTENT(IN)  :: PCLDFR  ! Cloud fraction
!
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PTHT    ! Theta at time t
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRVT    ! Water vapor m.r. at t 
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRCT    ! Cloud water m.r. at t 
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRRT    ! Rain water m.r. at t 
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PPABST  ! absolute pressure at t
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PTHS    ! Theta source
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRVS    ! Water vapor m.r. source
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRCS    ! Cloud water m.r. source
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRRS    ! Rain water m.r. source
!
REAL, DIMENSION(:,:), INTENT(INOUT)     :: PINPRR  ! Rain instant precip
REAL, DIMENSION(:,:,:), INTENT(INOUT)   :: PINPRR3D! Rain inst precip 3D
REAL, DIMENSION(:,:,:), INTENT(INOUT)   :: PEVAP3D ! Rain evap profil
!
END SUBROUTINE SLOW_TERMS
END INTERFACE
END MODULE MODI_SLOW_TERMS 
!
!
!     ######################################################################
!OPTION! -Ni
      SUBROUTINE SLOW_TERMS ( KSPLITR, PTSTEP, KMI, HSUBG_AUCV,            &
                              PZZ, PRHODJ, PRHODREF, PCLDFR,               &
                              PTHT, PRVT, PRCT, PRRT, PPABST,              &
                              PTHS, PRVS, PRCS, PRRS, PINPRR,              &
                              PINPRR3D, PEVAP3D                            )
!     ######################################################################
!
!!****  * -  compute the explicit microphysical sources 
!!
!!    PURPOSE
!!    -------
!!      The purpose of this routine is to compute the slow microphysical sources
!!    which can be computed explicitly 
!!
!!
!!**  METHOD
!!    ------
!!      The autoconversion computation follows Kessler (1969). 
!!      The sedimentation rate is computed with a time spliting technique and 
!!    an upstream scheme, written as a difference of non-advective fluxes. This
!!    source term is added to the future instant ( split-implicit process ).
!!      The others microphysical processes are evaluated at the central instant 
!!    (split-explicit process ): autoconversion, accretion and rain evaporation.
!!      These last 3 terms are bounded in order not to create negative values 
!!    for the water species at the future instant.  
!!
!!    EXTERNAL
!!    --------
!!      None
!!     
!!
!!    IMPLICIT ARGUMENTS
!!    ------------------
!!      Module MODD_PARAMETERS
!!          JPHEXT       : Horizontal external points number
!!          JPVEXT       : Vertical external points number
!!      Module MODD_CONF :
!!          CCONF configuration of the model for the first time step
!!      Module MODD_CLOUDPAR
!!          XCEXVT               ! constant in the rain drop fall velocity
!!          XC1RC, XC2RC         ! constants for autoconversion
!!          XCEXRA, XCRA         ! constants for accretion     
!!          XCEXRE, XC1RE, XC2RE ! constants for rain evaporation 
!!          XCEXRS, XCRS         ! constants for rain sedimentation
!!          XDIVA                ! vapor diffusivity in air
!!          XTHCO                ! thermal conductivity 
!!      Module MODD_CST     
!!          XP00               ! Reference pressure
!!          XRD,XRV            ! Gaz  constant for dry air, vapor
!!          XCPD               ! Cpd (dry air)
!!          XCL                ! Cl (liquid)
!!          XTT                ! Triple point temperature
!!          XLVTT              ! Vaporization heat constant
!!          XALPW,XBETAW,XGAMW ! Constants for saturation vapor 
!!                             !  pressure  function 
!!      Module MODD_BUDGET:
!!         NBUMOD       : model in which budget is calculated
!!         CBUTYPE      : type of desired budget
!!                          'CART' for cartesian box configuration
!!                          'MASK' for budget zone defined by a mask 
!!                          'NONE'  ' for no budget
!!         NBUPROCCTR   : process counter used for each budget variable
!!         LBU_RTH      : logical for budget of RTH (potential temperature)
!!                        .TRUE. = budget of RTH        
!!                        .FALSE. = no budget of RTH
!!         LBU_RRV      : logical for budget of RRV (water vapor)
!!                        .TRUE. = budget of RRV 
!!                        .FALSE. = no budget of RRV 
!!         LBU_RRC      : logical for budget of RRC (cloud water)
!!                        .TRUE. = budget of RRC 
!!                        .FALSE. = no budget of RRC 
!!         LBU_RRR      : logical for budget of RRR (rain water)
!!                        .TRUE. = budget of RRR 
!!                        .FALSE. = no budget of RRR 
!!
!!    REFERENCE
!!    ---------
!!
!!      Book1 and Book2 of documentation ( routine SLOW_TERMS )
!!
!!    AUTHOR
!!    ------
!!      E. Richard       * Laboratoire d'Aerologie*
!!
!!    MODIFICATIONS
!!    -------------
!!      Original    22/12/94 
!!      Modifications: June 8, 1995 ( J.Stein )
!!                                   Cleaning to improve efficienty and clarity
!!                                  in agreement with the MESO-NH coding norm
!!                     16/10/95 (J. Stein)     change the budget calls
!!                     16/10/96 (J. Stein)     use PEXNT instead of PEXNREF 
!!                     19/12/96 (J.-P. Pinty)  update the budget calls
!!                     04/02/97 (J.Viviand)    convert precipitation rate in m/s
!!                                             & debug accumulated precipitation
!!                     01/09/97 (V. Masson & J. Stein) optimization: reduction
!!                                             of the number of exponentiations
!!                     14/09/97 (V. Masson) removes low rr non-physical values
!!                     06/11/02 (V. Masson) update the budget calls
!!     J.Escobar : 15/09/2015 : WENO5 & JPHEXT <> 1
!  P. Wautelet 26/04/2019: replace non-standard FLOAT function by REAL function
!-------------------------------------------------------------------------------
!
!*       0.    DECLARATIONS
!              ------------
!
USE MODD_BUDGET,     only: LBUDGET_RC, LBUDGET_RR, LBUDGET_RV, LBUDGET_TH
USE MODD_CLOUDPAR,   only: XC1RC, XC2RC, XC1RE, XC2RE, XCEXRA, XCEXRE, XCEXRS, XCEXVT, XCRA, XCRS, XDIVA, XTHCO
USE MODD_CST,        only: XALPW, XBETAW, XGAMW, XCL, XCPD, XCPV, XLVTT, XMD, XMV, XP00, XRD, XRHOLW, XRV, XTT
USE MODD_PARAMETERS, only: JPVEXT

use mode_mppdb

#ifdef MNH_BITREP
use modi_bitrep
#endif
USE MODI_BUDGET
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
!
INTEGER,                  INTENT(IN)    :: KSPLITR ! Number of small time step 
                                      ! integration for  rain sedimendation
REAL,                     INTENT(IN)    :: PTSTEP  ! Double Time step
                                                   ! (single if cold start)
INTEGER,                  INTENT(IN)    :: KMI     ! Model index 
CHARACTER(LEN=4),         INTENT(IN)    :: HSUBG_AUCV
                                        ! Kind of Subgrid autoconversion method
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PZZ     ! Height (z)
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHODJ  ! Dry density * Jacobian
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHODREF! Reference density
REAL, DIMENSION(:,:,:),     INTENT(IN)  :: PCLDFR  ! Cloud fraction
!
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PTHT    ! Theta at time t
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRVT    ! Water vapor m.r. at t 
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRCT    ! Cloud water m.r. at t 
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRRT    ! Rain water m.r. at t 
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PPABST  ! absolute pressure at t
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PTHS    ! Theta source
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRVS    ! Water vapor m.r. source
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRCS    ! Cloud water m.r. source
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRRS    ! Rain water m.r. source
!
REAL, DIMENSION(:,:), INTENT(INOUT)     :: PINPRR  ! Rain instant precip
REAL, DIMENSION(:,:,:), INTENT(INOUT)   :: PINPRR3D! Rain inst precip 3D
REAL, DIMENSION(:,:,:), INTENT(INOUT)   :: PEVAP3D ! Rain evap profil
!
!*       0.2   Declarations of local variables :
!
INTEGER :: JK            ! Vertical loop index for the rain sedimentation 
INTEGER :: JN            ! Temporal loop index for the rain sedimentation
INTEGER :: IKB           !  Define the domain where is 
INTEGER :: IKE           !  the microphysical sources have to be computed
!
REAL    :: ZTSPLITR      ! Small time step for rain sedimentation
!
REAL,    DIMENSION(:,:,:), ALLOCATABLE :: ZT,ZW,ZW1,ZW2,ZW3,ZEXNT,ZDZZ  ! Work arrays
LOGICAL, DIMENSION(:,:,:), ALLOCATABLE :: G3D
!
INTEGER                              :: JI,JJ,IC,JL ! loop control for packed array

!-------------------------------------------------------------------------------

!$acc data present( PZZ, PRHODJ, PRHODREF, PCLDFR, PTHT, PRVT, PRCT, PRRT,      &
!$acc &             PPABST, PTHS, PRVS, PRCS, PRRS, PINPRR, PINPRR3D, PEVAP3D )

! !$acc &     copyin( XC1RC, XC2RC, XC1RE, XC2RE, XCEXRA, XCEXRE, XCEXRS, XCEXVT, XCRA, XCRS, XDIVA, XTHCO, &
! !$acc &             XALPW, XBETAW, XGAMW, XCL, XCPD, XCPV, XLVTT, XMD, XMV, XP00, XRD, XRHOLW, XRV, XTT )

IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PZZ,     "SLOW_TERMS beg:PZZ")
  CALL MPPDB_CHECK(PRHODJ,  "SLOW_TERMS beg:PRHODJ")
  CALL MPPDB_CHECK(PRHODREF,"SLOW_TERMS beg:PRHODREF")
  CALL MPPDB_CHECK(PCLDFR,  "SLOW_TERMS beg:PCLDFR")
  CALL MPPDB_CHECK(PTHT,    "SLOW_TERMS beg:PTHT")
  CALL MPPDB_CHECK(PRVT,    "SLOW_TERMS beg:PRVT")
  CALL MPPDB_CHECK(PRCT,    "SLOW_TERMS beg:PRCT")
  CALL MPPDB_CHECK(PRRT,    "SLOW_TERMS begPRRT:")
  CALL MPPDB_CHECK(PPABST,  "SLOW_TERMS beg:PPABST")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PTHS,    "SLOW_TERMS beg:PTHS")
  CALL MPPDB_CHECK(PRVS,    "SLOW_TERMS beg:PRVS")
  CALL MPPDB_CHECK(PRCS,    "SLOW_TERMS beg:PRCS")
  CALL MPPDB_CHECK(PRRS,    "SLOW_TERMS beg:PRRS")
  CALL MPPDB_CHECK(PINPRR,  "SLOW_TERMS beg:PINPRR")
  CALL MPPDB_CHECK(PINPRR3D,"SLOW_TERMS beg:PINPRR3D")
  CALL MPPDB_CHECK(PEVAP3D, "SLOW_TERMS beg:PEVAP3D")
END IF

allocate( zt   ( size(pzz,1), size(pzz,2), size(pzz,3) ) )
allocate( zw   ( size(pzz,1), size(pzz,2), size(pzz,3) ) )
allocate( zw1  ( size(pzz,1), size(pzz,2), size(pzz,3) ) )
allocate( zw2  ( size(pzz,1), size(pzz,2), size(pzz,3) ) )
allocate( zw3  ( size(pzz,1), size(pzz,2), size(pzz,3) ) )
allocate( zexnt( size(pzz,1), size(pzz,2), size(pzz,3) ) )
allocate( zdzz ( size(pzz,1), size(pzz,2), size(pzz,3) ) )
allocate( g3d   ( size(prhodj,1), size(prhodj,2), size(prhodj,3) ) )

!$acc data create( zt, zw, zw1, zw2, zw3, zexnt, zdzz, g3d )
!
!*       1.     COMPUTE THE LOOP BOUNDS AND EXNER FUNCTION
!   	        ------------------------------------------
!
IKB=1+JPVEXT
IKE=SIZE(PZZ,3) - JPVEXT
! compute Delta z at the mass point
!$acc kernels
DO JK = 1,IKE
  ZDZZ(:,:,JK) =  PZZ(:,:,JK+1)-PZZ(:,:,JK) 
END DO
!
!-------------------------------------------------------------------------------
!
!*       2.     COMPUTE THE SEDIMENTATION (RS) SOURCE
!	        -------------------------------------
!
!
!*       2.1    time splitting loop initialization        
!
ZTSPLITR = PTSTEP / REAL(KSPLITR)       ! Small time step
! 
ZW1(:,:,:) = PRRS(:,:,:) * PTSTEP
ZW2(:,:,:) = 0.
ZW3(:,:,:) = 0.
!
#ifndef MNH_OPENACC
G3D(:,:,:)=.FALSE.
DO JK=IKE,1,-1
  WHERE( ZW1(:,:,JK)>0. .OR. G3D(:,:,JK+1) )
    G3D(:,:,JK)=.TRUE.
  END WHERE
END DO
#else
g3d(:, :, : ) = .false.
do jk = ike, 1, -1
  do jj = 1, size(pzz,2)
    do ji = 1, size(pzz,1)
      if ( zw1(ji, jj, jk ) >0. .or. g3d(ji, jj, jk+1 ) ) then
        g3d(ji, jj, jk ) = .true.
      end if
    end do
  end do
end do
#endif
!
WHERE (G3D(:,:,:))
#ifndef MNH_BITREP
  ZW3(:,:,:) = PRHODREF(:,:,:) ** (XCEXRS-XCEXVT)
#else
  ZW3(:,:,:) = BR_POW( PRHODREF(:,:,:), XCEXRS - XCEXVT )
#endif
END WHERE
!
WHERE (ZW1(:,:,IKE+1)>0.)
#ifndef MNH_BITREP
  ZW2(:,:,IKE+1) =   XCRS                         &
                  * ZW1(:,:,IKE+1) ** XCEXRS      &
                  * PRHODREF(:,:,IKE+1) ** (XCEXRS-XCEXVT)
#else
  ZW2(:,:,IKE+1) =  XCRS                                          &
                  * BR_POW( ZW1(:,:,IKE+1), XCEXRS )              &
                  * BR_POW( PRHODREF(:,:,IKE+1), XCEXRS - XCEXVT )
#endif
END WHERE
!
!
!*       2.2    small time step integration
!               ---------------------------
!
DO JN=1,KSPLITR
!
!*       2.2.1    test where computations should be made and pack arrays
!                 ------------------------------------------------------
!

  DO JK=IKE,IKB,-1
!$acc loop collapse(2) independent
    DO JJ = 1,SIZE( ZW1,2)
      DO JI = 1,SIZE( ZW1,1)
        IF ( ( ZW1(JI,JJ,JK+1)>0. ) .AND. ( ZW1(JI,JJ,JK)>0. ) )  THEN
!
!*       2.2.2    compute the rain flux
!                 ---------------------
!
!
#ifndef MNH_BITREP
          ZW2(JI,JJ,JK) = XCRS * ( ZW1(JI,JJ,JK) ** XCEXRS ) * ZW3(JI,JJ,JK)
#else
          ZW2(JI,JJ,JK) = XCRS * BR_POW( ZW1(JI,JJ,JK), XCEXRS ) * ZW3(JI,JJ,JK)
#endif
!
!
!*       2.2.3    update the rain tendency with the small timestep
!                 ------------------------------------------------
!
          ZW1(JI,JJ,JK)  = ZW1(JI,JJ,JK) + ZTSPLITR *       &
                    ( ZW2(JI,JJ,JK+1)-ZW2(JI,JJ,JK) ) /    &
                    ( ZDZZ(JI,JJ,JK) * PRHODREF(JI,JJ,JK) )
        END IF
      END DO
    ENDDO
  ENDDO
!
!*       2.2.4    compute the explicit accumulated precipitations
!                 -----------------------------------------------
! 
  IF( JN==1 ) THEN
    PINPRR(:,:) = ZW2(:,:,IKB)/XRHOLW                           ! in m/s
    PINPRR3D(:,:,:) = ZW2(:,:,:)/XRHOLW                           ! in m/s
  END IF
!
END DO
!
!*       2.4     update the rain tendency
!
PRRS(:,:,:) = ZW1(:,:,:) / PTSTEP
!$acc end kernels
!
!*       2.5     budget storage
!
IF (LBUDGET_RR) THEN
!$acc update self(PRRS)
  CALL BUDGET (PRRS(:,:,:)*PRHODJ(:,:,:),8,'SEDI_BU_RRR')
END IF
!
!-------------------------------------------------------------------------------
!
!
!*       3.     COMPUTES THE ACCRETION SOURCE
!   	        -----------------------------
!
!
!*       3.1     compute the accretion and update the tendencies
!
!$acc kernels
G3D(:,:,:) = PRCT(:,:,:)>0.0 .AND. PRRT(:,:,:)>0.0 .AND. PRCS(:,:,:)>0.0
WHERE ( G3D(:,:,:) )
#ifndef MNH_BITREP
  ZW(:,:,:) = XCRA * PRCT(:,:,:)                        &
             * PRRT(:,:,:) ** XCEXRA                    &
             * PRHODREF(:,:,:) ** (XCEXRA - XCEXVT)
#else
  ZW(:,:,:) = XCRA * PRCT(:,:,:)                         &
             * BR_POW( PRRT(:,:,:), XCEXRA )             &
             * BR_POW( PRHODREF(:,:,:), XCEXRA - XCEXVT )
#endif
  ZW(:,:,:) = MIN ( ZW(:,:,:) , PRCS(:,:,:) )
!
  PRCS(:,:,:) = PRCS(:,:,:) - ZW(:,:,:)
  PRRS(:,:,:) = PRRS(:,:,:) + ZW(:,:,:)
END WHERE
!$acc end kernels
!
!*       3.2     budget storage
!
IF (LBUDGET_RC) THEN
!$acc update self(PRCS)
  CALL BUDGET (PRCS(:,:,:)*PRHODJ(:,:,:),7,'ACCR_BU_RRC')
END IF
IF (LBUDGET_RR) THEN
!$acc update self(PRRS)
  CALL BUDGET (PRRS(:,:,:)*PRHODJ(:,:,:),8,'ACCR_BU_RRR')
END IF
!
!-------------------------------------------------------------------------------
!
!
!*       4.     COMPUTES THE AUTOCONVERSION SOURCE
!               ----------------------------------
!
!
!*       4.1     compute the autoconversion and update the tendencies
!
!$acc kernels
IF ( HSUBG_AUCV == 'CLFR' ) THEN
 G3D(:,:,:) = PRCT(:,:,:)>0.0 .AND. PRCS(:,:,:)>0.0 .AND. PCLDFR(:,:,:)>0.0
 WHERE ( G3D(:,:,:) )
  ZW(:,:,:) = XC1RC * MAX(PRCT(:,:,:) /(PCLDFR(:,:,:)) - XC2RC / PRHODREF(:,:,:),0.)
  ZW(:,:,:) = MIN ( (ZW(:,:,:)* PCLDFR(:,:,:)), PRCS(:,:,:) )
  PRCS(:,:,:) = PRCS(:,:,:) - ZW(:,:,:)
  PRRS(:,:,:) = PRRS(:,:,:) + ZW(:,:,:)
 END WHERE
ELSE
 G3D(:,:,:) = PRCT(:,:,:)>0.0 .AND. PRCS(:,:,:)>0.0
 WHERE ( G3D(:,:,:) )
  ZW(:,:,:) = XC1RC * MAX ( PRCT(:,:,:) - XC2RC / PRHODREF(:,:,:), 0. )
  ZW(:,:,:) = MIN ( ZW(:,:,:) , PRCS(:,:,:) )
!
  PRCS(:,:,:) = PRCS(:,:,:) - ZW(:,:,:)
  PRRS(:,:,:) = PRRS(:,:,:) + ZW(:,:,:)
 END WHERE
END IF
!$acc end kernels
!
!*       4.2     budget storage
!
IF (LBUDGET_RC) THEN
!$acc update self(PRCS)
  CALL BUDGET (PRCS(:,:,:)*PRHODJ(:,:,:),7,'AUTO_BU_RRC')
END IF
IF (LBUDGET_RR) THEN
!$acc update self(PRRS)
  CALL BUDGET (PRRS(:,:,:)*PRHODJ(:,:,:),8,'AUTO_BU_RRR')
END IF
!
!-------------------------------------------------------------------------------
!
!*       5.     COMPUTES THE RAIN EVAPORATION (RE) SOURCE
!   	        -----------------------------------------
!
!$acc kernels
PEVAP3D(:,:,:)=0.
G3D(:,:,:) = PRRT(:,:,:)>0.0 .AND. PRCT(:,:,:)==0.0
WHERE ( G3D(:,:,:) )
!
!*       5.1    compute the Exner function
!
#ifndef MNH_BITREP
  ZEXNT(:,:,:) = (PPABST(:,:,:)/XP00)**(XRD/XCPD)
#else
  ZEXNT(:,:,:) = BR_POW( PPABST(:,:,:)/XP00, XRD/XCPD )
#endif
!
!*       5.2    compute the temperature
!
  ZT(:,:,:) = PTHT(:,:,:) * ZEXNT(:,:,:)
!
!*       5.3    compute the saturation vapor pressure
!
#ifndef MNH_BITREP
  ZW1(:,:,:) = EXP( XALPW - XBETAW/ZT(:,:,:) - XGAMW*ALOG(ZT(:,:,:) ) )
#else
  ZW1(:,:,:) = BR_EXP( XALPW - XBETAW/ZT(:,:,:) - XGAMW*BR_LOG(ZT(:,:,:) ) )
#endif
!
!*       5.4    compute the saturation mixing ratio
!
  ZW2(:,:,:) = (XMV/XMD) * ZW1(:,:,:)                &
                         / ( PPABST(:,:,:) - ZW1(:,:,:) )  
!
!*       5.5    compute the latent heat of vaporization
!
  ZW3(:,:,:) = XLVTT + ( XCPV - XCL ) * ( ZT(:,:,:) -XTT )
!
!*       5.6    compute the source
!
#ifndef MNH_BITREP
  ZW(:,:,:) = MAX( 1. - PRVT(:,:,:)/ZW2(:,:,:) , 0.0 ) *         &
    ( XC1RE * SQRT( PRRT(:,:,:)/PRHODREF(:,:,:) )                & 
     +XC2RE * PRRT(:,:,:)**XCEXRE                                & 
            * PRHODREF(:,:,:)**(XCEXRE-0.5*XCEXVT-1.)            &
    ) /                                                          &
    ( XRV*ZT(:,:,:)/(ZW1(:,:,:)*XDIVA)                           & 
     +ZW3(:,:,:) ** 2 / (XTHCO*XRV*ZT(:,:,:)**2)                 &
    )
#else
  ZW(:,:,:) = MAX( 1. - PRVT(:,:,:)/ZW2(:,:,:) , 0.0 ) *         &
    ( XC1RE * BR_POW( PRRT(:,:,:)/PRHODREF(:,:,:), 0.5 )         &
     +XC2RE * BR_POW( PRRT(:,:,:), XCEXRE )                      &
            * BR_POW( PRHODREF(:,:,:), XCEXRE-0.5*XCEXVT-1. )    &
    ) /                                                          &
    ( XRV*ZT(:,:,:)/(ZW1(:,:,:)*XDIVA)                           &
     +ZW3(:,:,:)*ZW3(:,:,:) / (XTHCO*XRV*ZT(:,:,:)*ZT(:,:,:))    &
    )
#endif
  ZW(:,:,:) = MIN ( ZW(:,:,:) , PRRS(:,:,:) )
!
!*       5.7     update the total tendencies
!
  PRRS(:,:,:) = PRRS(:,:,:) - ZW(:,:,:)
  PRVS(:,:,:) = PRVS(:,:,:) + ZW(:,:,:)
!
  PTHS (:,:,:) = PTHS (:,:,:)  - ZW(:,:,:) * ZW3(:,:,:) /  &
    ( ZEXNT(:,:,:) * (  XCPD + XCPV * PRVT(:,:,:)          &
                      + XCL * (PRCT(:,:,:) + PRRT(:,:,:)) ) )
  PEVAP3D(:,:,:)=ZW(:,:,:)
END WHERE
!$acc end kernels
!
!*       5.8     budget storage
!
IF (LBUDGET_RV) THEN
!$acc update self(PRVS)
  CALL BUDGET (PRVS(:,:,:)*PRHODJ(:,:,:),6,'REVA_BU_RRV')
END IF
IF (LBUDGET_RR) THEN
!$acc update self(PRRS)
  CALL BUDGET (PRRS(:,:,:)*PRHODJ(:,:,:),8,'REVA_BU_RRR')
END IF
IF (LBUDGET_TH) THEN
!$acc update self(PTHS)
  CALL BUDGET (PTHS(:,:,:)*PRHODJ(:,:,:),4,'REVA_BU_RTH')
END IF
!
!-------------------------------------------------------------------------------
!
!*       6.     REMOVES NON-PHYSICAL LOW VALUES
!   	        -------------------------------
!
!$acc kernels
G3D(:,:,:) = PRRS(:,:,:) < 1.E-16
WHERE ( G3D(:,:,:) )
  PRRS(:,:,:)=0.
END WHERE
!$acc end kernels
!-------------------------------------------------------------------------------
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PTHS,    "SLOW_TERMS end:PTHS")
  CALL MPPDB_CHECK(PRVS,    "SLOW_TERMS end:PRVS")
  CALL MPPDB_CHECK(PRCS,    "SLOW_TERMS end:PRCS")
  CALL MPPDB_CHECK(PRRS,    "SLOW_TERMS end:PRRS")
  CALL MPPDB_CHECK(PINPRR,  "SLOW_TERMS end:PINPRR")
  CALL MPPDB_CHECK(PINPRR3D,"SLOW_TERMS end:PINPRR3D")
  CALL MPPDB_CHECK(PEVAP3D, "SLOW_TERMS end:PEVAP3D")
END IF

!$acc end data

!$acc end data

END SUBROUTINE SLOW_TERMS
